#!/usr/bin/python3
#-*- encoding: utf-8 -*-
import os, stat
import locale
import sys
import hivex
import json
import gettext
import re
from enum import Enum
from operator import *

# i18n
gettext.bindtextdomain("gvfs", "/usr/share/locale")
gettext.textdomain("gvfs")
__ = gettext.gettext

#fat32的分区类型是vfat
WinDiskTypes = ['exfat', 'vfat', 'ntfs']
# Windows系统分区类型：[Microsoft basic data(gpt), gdisk hex code(dos)]
WinSysPartTypes = ['ebd0a0a2-b9e5-4433-87c0-68b6b72699c7', '0x7']
LinuxDiskTypes = ['ext2', 'ext3', 'ext4']
# Linux系统分区类型：[Linux filesystem(gpt), gdisk hex code(dos)]
LinuxSysPartTypes = ['0fc63daf-8483-4772-8e79-3d69d8477de4', '0x83']

#windows系统注册表所在目录，当前暂不考虑windows装载在别的目录的情况
WindowsRegistryPath = 'Windows/System32/config'
WindowsRegistryPathXP = 'WINDOWS/system32/config'

#windows盘符对应驱动器枚举情况，因为手边没有合适设备测试，暂时没有考虑软驱、以及zip disk的情况，后面碰到对应实际场景时再加上
class WinDriverKindsEnum(Enum):
    KIND_FIXED_DISK = 0
    KIND_DMIO = 1
    KIND_USBSTORE_CDROM = 2
    KIND_IDE_CDROM = 3
    KIND_SCSI_CDROM = 4
    KIND_USBSTORE_DISK = 5
    KIND_OTHER = 6

LANGUAGE=locale.getdefaultlocale()[0]
def _(s):
    interpretor = {'Device':u'设备', 'Disk identifier':u'磁盘标识符', ':':u'：', 'Disk Label':'磁盘标签类型'}
    if eq (LANGUAGE, 'en_US'):
        return s
    else:
        return interpretor[s]

#函数返回windows设备信息列表，以及物理磁盘信息列表
#对fdisk -l返回的信息进行分析，返回各个windows分区的driver和起点
#函数中我们以'设备'和'磁盘标识符'作为定位字符串
#英文环境通过自定义的 _() 函数支持
def listWindowsDevicesAndPhysicalDisks():
    res = os.popen('sudo fdisk -l').read()
    #flist -l信息分成多段，段间以空行分割
    paras = res.split('\n\n\n')
    #if gt(len(paraDevices), 1):
    #    otherDevices = paraDevices[1]
    #else:
    #    otherDevices = None
    #if otherDevices is not None:
    #    otherLineDevices = otherDevices.split('\n')[1:]
    #    otherDevices = [line.split() for line in otherLineDevices]
    #FIXME 多块硬盘时，第二块盘上的 deviceLinus 自动挂载?
    #下面len(words)==8代表设备是个启动设备(‘启动'列为*，非启动设备'启动'列为空格或\t

    deviceWins={}
    deviceLinus={}
    diskPhysicals={}

    for para in paras:
        #找到含有磁盘标识符的那一段，找到倒数第二行，也就是标签为 dos/gpt
        #有一个包含磁盘信息的信息，这个段以'设备\t启动\t起点\t末尾\t扇区\t大小\tId\t类型'行开头，下面的数据是个表形式
        lparas = para.split('\n\n')
        print (para)
        if eq(LANGUAGE, "en_US"):
            paraType = [identifier.split(_(': ')) for identifier in  para.split('\n') if identifier.startswith(_('Disklabel type'))]
        elif eq (LANGUAGE, 'zh_CN'):
            paraType = [identifier.split(_(':')) for identifier in  para.split('\n') if identifier.startswith(_('Disk Label'))]
        else:
            print ("language is not English or Chinese")
        if len(paraType) == 0:
            continue

        #for section in lparas:
        #    if section.startswith(_('Device')):
        #        print (section)
        lineDevices = [section.split('\n') for section in lparas if section.startswith(_('Device'))]
        if len(lineDevices) == 0:
            continue
        deleteHead = lineDevices[0][1:]
        devices=[]
        for line in deleteHead:
            devices.append (line.split())

        #对于每块设备找到对应的 windows 设备和 linus 设备
        if eq(paraType[0][-1], "dos"):
            device = para.split(_(':'))[0].split()[1]
            deviceWins.update({words[0] : (int(words[2]) if len(words)==8 else int(words[1])) for words in devices if len(words)!=0 and isWindowsSystemPartition(words[0])})
            deviceLinus.update({words[0] : (int(words[2]) if len(words)==8 else int(words[1])) for words in devices if len(words)!=0 and isLinuxSystemPartition(words[0]) and not isRootOrHome(words[0])})
            diskPhysicals.update({device : int(item.split(_(':'))[-1].strip()[2:],16) for item in para.split('\n') if item.startswith(_('Disk identifier'))})
        elif eq(paraType[0][-1], "gpt"):
            if eq(LANGUAGE, "en_US"):
                index = 5
            else:
                index = -2
        # 第一块 gpt 分区的设备
            deviceWins.update({words[0] : (int(words[2]) if len(words)==8 else int(words[1])) for words in devices if len(words)!=0 and isWindowsSystemPartition(words[0])})
            deviceLinus.update({words[0] : (int(words[2]) if len(words)==8 else int(words[1])) for words in devices if len(words)!=0 and isLinuxSystemPartition(words[0]) and not isRootOrHome(words[0])})
            #guid = [para.split('\n')[-1].split(_(':'))[-1] for para in paras if para.split('\n')[-1].startswith(_('Disk identifier'))][0].split(_(':'))[-1]
            diskPhysicals.update({"device":"guid"})
        else:
            print("error")

    print ("----------------------------------------------------")
    print (deviceWins)
    print (deviceLinus)
    return deviceWins, deviceLinus, diskPhysicals

#执行mount，列出所有设备mount信息,我们关心的是设备以及所mount的目录两个内容
#mount 有问题，如果 windows 重命名有空格，比如 This is C
#不太方便找路径，改为 json 解析，方便
def listDeviceMounts(device):
    resJson = os.popen('findmnt -J --source %s' % device).read()
    if len(resJson) == 0:
        return None

    value = json.loads(resJson)
    return value['filesystems'][0]['target']

#获得设备挂载缺省目录
def getPathMount(device):
    path = '/media/nfs_dev_%s' % device.split('/')[-1]
    #print ('getPathMount: %s - %s' % (device, path))
    return path

def getDeviceByLabel(label):
    if label == None:
        return None

    res = os.popen('blkid -L %s' % label).read()
    res = res.strip('\n').strip()
    return res

def setDeviceLabel(device):
    res = os.popen('e2label %s' % device).read()
    res = res.strip('\n').strip()
    if len(res) == 0:
        os.system("e2label %s '%s'" % (device, __("Local Volume")))

#跳过不自动挂载的分区
def checkPartMount(device):
    #Lookup options of /etc/fstab. 
    device_u = os.popen('blkid %s -s UUID' % device).read()
    uuid = device_u.split('=')[-1]
    uuid = uuid.replace('"','').strip('\n').strip()
    #print ('uuid = %s' % (uuid))
    fstab_f = open("/etc/fstab")
    need_mount = True
    for line in fstab_f:
        line = line.strip('\n').strip()
        if line[0] == '#':
            continue
        fields = line.split(' ')
        if fields[0].split('=')[-1] == uuid:
            #print (fields)
            x_gvfs_show = False
            noauto = False
            for conf in fields:
                if not conf:
                    continue
                if 'x-gvfs-show' in conf:
                    x_gvfs_show = True
                if 'noauto' in conf:
                    noauto = True
            #不显示，且不自动挂载，就不挂载,如果需要显示，就自动挂载。
            if x_gvfs_show == False or noauto:
                need_mount = False
    fstab_f.close()
    return need_mount

#挂载设备
def doDeviceMount(device):
    res = os.popen('blkid %s -s TYPE' % device).read()
    
    if len(res) == 0 or "swap" in res:
        return None

    setDeviceLabel(device)
    pathMount = getPathMount(device)
    if not os.path.exists(pathMount):
        res = os.system('mkdir %s' % pathMount)
        if res != 0:
            print('Cannot mkdir path %s' % pathMount)
            return None
    
    res = os.system('mount %s %s' % (device, pathMount))
    if res != 0:
        print ('Cannot mount %s %s with result=%d' % (device, pathMount, res))
        return None
    os.chmod(pathMount, stat.S_IRWXO | stat.S_IRWXG | stat.S_IRWXU)
    return pathMount

#装载注册表文件
def loadRegistry(file):
    registry = hivex.Hivex(file)
    return registry

#获得注册表文件中某个键值对应节点，该键值是个路径，以/分割
def getNodeByPath(registry, keyPaths):
    keys = keyPaths.split('/')
    current = registry.root()
    i = 0
    while i < len(keys):
        key = keys[i]
        current = registry.node_get_child(current, key)
        if current is None:
            print('key %s not found!' % '/'.join(keys[:i+1]))
            return None
        i += 1
    return current

#bytes字节倒序对应的十进制数
def to_hex(bstr):
    return sum([bstr[i]<<(8*i) for i in range(len(bstr))])

#获得windows盘符信息，包括是固定磁盘，还是U盘、CD等
#固定磁盘信息共12个字节，前4个字节倒序是该磁盘的id，后4个字节倒序是该磁盘在分区的偏移值
#固定磁盘信息共24个字节，说明是 gpt 格式，前8个字节顺序是 DMIP:ID:，后16个字节是 GUID
#GUID 格式：xxxx-xx-xx-xx-xxxxxx [后两组同文本序]
#注册表读出来 b'\xdc2\x85\x19\xfd\x1c\x8bD\x87\x89\x9cp\xb9n\x8fv'
#其他磁盘待后续有真实测试设备，再分类
def getDriverKindAndInfo(bstr):
    if len(bstr)==12: 
        #固定磁盘
        idDriver = to_hex(bstr[:4])
        offsetDriver = to_hex(bstr[4:]) >> 9  
        return WinDriverKindsEnum.KIND_FIXED_DISK, [idDriver, offsetDriver]
    if len(bstr)==24:
        hexstring = bstr[8:].hex()
        bytesarr = ':'.join(hexstring[i:i+2] for i in range(0, len(hexstring), 2)).split(':')
        temp = bytesarr[0:8]
        temp.reverse()
        third = ''.join(temp[i] for i in range(0,2))
        second = ''.join(temp[i] for i in range(2, 4))
        first = ''.join(temp[i] for i in range(4, 8))
        forth = ''.join(bytesarr[i] for i in range (8, 10))
        fifth = ''.join(bytesarr[i] for i in range (10, 16))
        guid = '%s-%s-%s-%s-%s' % (first, second, third, forth, fifth)
        return WinDriverKindsEnum.KIND_DMIO, guid

    info = ''.join([chr(bstr[i]) for i in range(len(bstr)) if bstr[i]!=0])
    if info.startswith('\\??\\USBSTOR#CdRom'):
        kind = WinDriverKindsEnum.KIND_USBSTORE_CDROM
    elif info.startswith('\\??\\IDE#CdRom'):
        kind = WinDriverKindsEnum.KIND_IDE_CDROM
    elif info.startswith('\\??\\SCSI#CdRom'):
        kind = WinDriverKindsEnum.KIND_SCSI_CDROM
    elif info.startswith('_??_USBSTOR#Disk'):
        kind = WinDriverKindsEnum.KIND_USBSTORE_DISK
    elif info.startswith('DMIO:ID:'):
        kind = WinDriverKindsEnum.KIND_DMIO
    else:
        #以后遇到other kind，需要认真分析
        print('other kind: %s' % info)
        kind = WinDriverKindsEnum.KIND_OTHER
    return kind, info

#读注册表SYSTEM/MountedDevices信息
#所有分配过的盘符，都在此注册项中，以\\DosDevice\开头, like '\\Device\C:'
#这里我们没有测试过windows用subst命令映射的盘符
def listDriverWins(registry):
    root = registry.root()
    nodeMountedDevices = registry.node_get_child(root, 'MountedDevices')
    values = registry.node_values(nodeMountedDevices)
    drivers = {registry.value_key(value) : registry.value_value(value)[1] for value in values}
    drivers = {key:getDriverKindAndInfo(drivers[key]) for key in drivers.keys() if key.startswith('\\DosDevices\\') and key.endswith(':')}
    return drivers

#读取注册表 SOFTWARE 的 Microsoft\Windows NT\CurrentVersion 的 RegisteredOwner
#windows用户名在这个注册表项中保存
#后来发现读的很不准，换从 sam 表读取
def loadUserName(registry):
    root = registry.root()
    #nodeCurrentVersion = registry.node_get_child(root, 'Microsoft/Windows NT/CurrentVersion')
    nodeUsers = getNodeByPath(registry, 'SAM/Domains/Account/Users/Names')
    childrens = registry.node_children(nodeUsers)
    names = [registry.node_name(child) for child in childrens]
    return names

#列出硬盘/IDE光盘所对应盘符,不包括移动光驱,硬盘是KIND_FIXED_DISK，光盘是KIND_IDE_CDROM或KIND_SCSI_CDROM
def listDriverWinsFixed(driverWins):
    #Windows分区不应该包含光驱
    drivers = {key.split('\\')[-1].split(':')[0] : driverWins[key] for key in driverWins if driverWins[key][0] in [ WinDriverKindsEnum.KIND_FIXED_DISK, WinDriverKindsEnum.KIND_DMIO]}
    return drivers

def findMatchDeviceGPT(guid):
    devicelink = "/dev/disk/by-partuuid/"+guid

    if os.path.exists(devicelink):
        return os.readlink(devicelink)
    return None

def checkWindowsRegistryFile(deviceWins, deviceMountsOfWin):
    fileSystemRegistry = None
    fileSamRegistry = None
    xp = False
    for device in deviceWins:
        folderName = deviceMountsOfWin[device] + '/' + WindowsRegistryPath 
        if not os.path.exists (folderName):
            folderName = deviceMountsOfWin[device] + '/' + WindowsRegistryPathXP
            if os.path.exists (folderName):
                xp = True
            else:
                continue
        
        print(folderName)
        fileSystemRegistry = folderName + '/SYSTEM'
        fileSystemRegistrybak = folderName + '/system'
        fileSamRegistry = folderName + '/SAM'
        fileSamRegistrybak = folderName + '/sam'

        if os.path.exists(fileSystemRegistrybak): 
            fileSystemRegistry = fileSystemRegistrybak
            
        if os.path.exists (fileSamRegistrybak): 
            fileSamRegistry = fileSamRegistrybak 

        if fileSystemRegistry is not None:
            return fileSystemRegistry, fileSamRegistry, xp 

    return fileSystemRegistry, fileSamRegistry, xp

#搜索物理磁盘列表(包含多个磁盘名称和id, 形式为[/dev/sda, 0xaabbccdd],
#和windows盘符信息列表(包含多个设备名称和offset，形式为[/dev/sda1, 0xeeffgghh],
#根据磁盘id信息[0xaabbccdd]，
#磁盘offset信息[0xeeffgghh],
#找到其对应设备为/dev/sda1 [0xaabbccdd -> disk /dev/sda, (/dev/sda1, 0xeeffgghh) in (/dev/sda, 0xeeffgghh)]
#找到对应设备：/dev/sda1
#===
#更新：gpt 格式保存的是 [/dev/sda, GUID]
#GUID: 16字节，不能对应到分区，所以需要换一种方式找到分区与 GUID 的映射
def findMatchDevice(diskPhysicals, deviceWins, idDriver, offsetDriver):
    #print('diskPhysicals=',diskPhysicals)
    #print('deviceWins=',deviceWins)
    #print('idDriver=',idDriver)
    #print('offsetDriver=', offsetDriver)
    if idDriver not in diskPhysicals.values():
        return None

    disk = [disk for disk in diskPhysicals if diskPhysicals[disk]==idDriver][0]
    #print('disk', disk)
    for device in deviceWins:
        #print('device', device)
        #print('deviceWins[device]', deviceWins[device], offsetDriver)
        #print (device.startswith(disk))
        #print(deviceWins[device], offsetDriver)
        if device.startswith(disk) and deviceWins[device]==offsetDriver:
            return device

    return None

def isInLivecd():
    f = open('/proc/cmdline', 'r')
    res = False

    for line in f.readlines():
        if 'boot=live' in line:
            res = True
            break

    f.close()

    return res

def getLabel(device):
    res = os.popen('lsblk -o label %s' % device).readlines()
    count = 0

    for line in res:
        count = count + 1
        line = line.strip('\n')
        if count == 2:
            return line

def isRootOrHome(device):
    res = False
    label = getLabel(device)
    if label == 'rootfs' or label == 'home':
        res = True
    
    return res

def getDeviceInfoFromLsblk(device):
    res = os.popen('lsblk -o parttype,fstype,hotplug %s' % device).readlines()
    count = 0

    for line in res:
        count = count + 1
        line = line.strip('\n')
        if count == 2:
            line = re.sub(r'\s+', ' ', line)
            resList = line.split(' ')

            if len(resList) == 2:
                resList.insert(1, '')

            resList[2] = int((resList[2])) == 1

            return resList

def isLinuxSystemPartition(device):
    partType, fsType, isHotPlug = getDeviceInfoFromLsblk(device)
    res = False

    if partType in LinuxSysPartTypes and fsType in LinuxDiskTypes and not isHotPlug:
        res = True
    
    return res

def isWindowsSystemPartition(device):
    partType, fsType, isHotPlug = getDeviceInfoFromLsblk(device)
    res = False

    if partType in WinSysPartTypes and fsType in WinDiskTypes and not isHotPlug:
        res = True
    
    return res

def doMatch():
    deviceWins, deviceLinus, diskPhysicals = listWindowsDevicesAndPhysicalDisks()
    #对于每个windows设备，如果已经被挂载，则记录其挂载目录，如果未被挂载，则执行doDeviceMount对其进行挂载在某个指定目录，并记录该指定目录
    for device in deviceLinus:
        if checkPartMount(device) == False:
            continue
        #livecd下不挂载Linux分区
        if isInLivecd():
            continue
        if listDeviceMounts(device) is None:
            doDeviceMount(device)

    deviceMountsOfWin = {}
    for device in deviceWins:
        value = listDeviceMounts(device)
        if value is not None:
            deviceMountsOfWin[device] = value
        else:
            pathMount = None
            if checkPartMount(device) == True:
                pathMount = doDeviceMount(device)
            if pathMount is not None:
                deviceMountsOfWin[device] = pathMount
            else:
                deviceMountsOfWin[device] = "safe"

    #we assue there only one windisk are windows system driver
    fileSystemRegistry, fileSamRegistry, xp = checkWindowsRegistryFile (deviceWins, deviceMountsOfWin)
    if fileSystemRegistry is None: 
        print('There have no windows system partition disk')
        sys.exit(0)

    #获得盘符对应信息
    print (fileSystemRegistry)
    registrySystem = loadRegistry(fileSystemRegistry)
    driverWins = listDriverWins(registrySystem)
    driverWinsFixed = listDriverWinsFixed(driverWins)
    print (driverWins)

    #获得 windows 的用户名
    registrySam = loadRegistry(fileSamRegistry)
    userName = loadUserName (registrySam)
    print(userName)
    #拼凑这个用户的桌面路径
    realDesktopPath = {}
    for device in deviceMountsOfWin:
        for user in userName:
            if not xp:
                DesktopPath = {user:deviceMountsOfWin[device]+'/Users/'+user+'/Desktop'}
            else:
                DesktopPath = {user:deviceMountsOfWin[device]+'/Documents and Settings/'+user+'/桌面'}
            
            if os.path.exists(DesktopPath[user]):
                realDesktopPath.update(DesktopPath)
    
    deviceOfWindowDrivers = {}
    for driver in driverWinsFixed:
        kind = driverWinsFixed[driver][0]
        if kind == WinDriverKindsEnum.KIND_FIXED_DISK:
            idDriver = driverWinsFixed[driver][1][0]
            offsetDriver = driverWinsFixed[driver][1][1]
            device = findMatchDevice(diskPhysicals, deviceWins, idDriver, offsetDriver)
            deviceOfWindowDrivers[driver] = device
        elif kind == WinDriverKindsEnum.KIND_DMIO:
            guid = driverWinsFixed[driver][1]
            device = findMatchDeviceGPT(guid) #FIXME
            gptdevice = '/dev/'+device.split('/')[-1]
            if device is not None:
                deviceOfWindowDrivers[driver] = gptdevice
            else:
                deviceOfWindowDrivers[driver] = "error"
        #Windows分区不应该包含光驱
        '''
        elif kind == WinDriverKindsEnum.KIND_IDE_CDROM:
            deviceOfWindowDrivers[driver] = 'Ide-Cdrom' 
        elif kind == WinDriverKindsEnum.KIND_SCSI_CDROM:
            deviceOfWindowDrivers[driver] = 'Scsi-Cdrom' 
        '''
    return deviceOfWindowDrivers, realDesktopPath
    
if __name__ == '__main__':
    deviceOfWindowDrivers, DesktopPath = doMatch()
    fp = open ('/etc/systemd/windows.ini', 'w')
    fp.write('[Windows]\n')

    print (deviceOfWindowDrivers)
    for device in deviceOfWindowDrivers:
        if deviceOfWindowDrivers[device] is not None:
            str = device+'='+deviceOfWindowDrivers[device]+'\n'
            fp.write(str)


    #if len(DesktopPath)!=0:
    #    fp.write('[Desktop]\n')
    #    for name in DesktopPath:
    #        desktopPath = name+'='+DesktopPath[name]+'\n'
    #        fp.write(desktopPath)

    fp.write('[MountPath]\n')
    for device in deviceOfWindowDrivers:
        if deviceOfWindowDrivers[device] is not None:
            mountpath = getPathMount (deviceOfWindowDrivers[device])
            if os.path.exists(mountpath):
                str = device+'='+mountpath+'\n'
                fp.write(str)
    fp.close()
